#include <netinet/in.h>
#include <netinet/ip.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>


#define MAXSOCKETS 128
#define TIMEOUT	 3

struct hostent *host_entry;

void portfound(int port)
{
    struct servent *service;
    
    printf("Open port : %d",port);
    service=getservbyport(htons(port),"tcp");
    if (service)
    {
	printf(" (%s)",service->s_name);
    }
    printf("\n");
}

void scanports(int start, int end)
{
    int *sock;
    int i,val,len;
    fd_set rset, wset;
    struct timeval tv;

    printf("Scanning ports %d-%d\n",start,end);
    len=sizeof(val);
    sock=(int *)malloc((end-start+1)*sizeof(int));

    for(i=start;i<=end;i++)
    {
	sock[i-start]=connect2port(i);
    }

    while(1) {
        tv.tv_sec = TIMEOUT;
        tv.tv_usec = 0;
        FD_ZERO(&rset);
        FD_ZERO(&wset);
	for(i=0;i<=end-start;i++) 
	    if (sock[i]!=0)
    	    {
    		FD_SET(sock[i],&rset);
		FD_SET(sock[i],&wset);
	    }
	//Ha nincs semmi, akkor ki
	if (select(FD_SETSIZE, &wset, &rset, NULL, &tv) < 1) break;
	
	for(i=0;i<=end-start;i++) 
	    if (sock[i]!=0)
	    {
		if (FD_ISSET(sock[i], &rset))
		{
		if (!getsockopt(sock[i], SOL_SOCKET, SO_ERROR, &val, &len)) 
		{
		    if (!val) {
    			portfound(i+start);
			close(sock[i]);
			sock[i]=0;
			continue;
		    }
		    else {
			close(sock[i]);
			sock[i]=0;
			continue;
		    }
		}
		}
	    }
	for(i=0;i<=end-start;i++) 
	    if (sock[i]!=0)
	    {
		if (FD_ISSET(sock[i], &wset))
		{
		    portfound(i+start);
		    close(sock[i]);
		    sock[i]=0;
		}
	    }
    }
    
    free(sock);
}

int connect2port(int port)
{
    struct sockaddr_in host_addr;
    int sd;
    int old_fcntl;
    struct linger l;

    l.l_onoff = 1;
    l.l_linger = 0;

    //TCP socket letrehozasa
    if ((sd=socket(AF_INET,SOCK_STREAM,6)) < 0)
    {
        perror("socket err");
        exit(1);
    }
    //nonblock
    old_fcntl = fcntl(sd, F_GETFL, 0);
    fcntl(sd, F_SETFL, old_fcntl | O_NONBLOCK);
    //linger
    if (setsockopt(sd, SOL_SOCKET, SO_LINGER,  (void *) &l, sizeof(struct linger)))
    {
	perror("setsockopt");
    }
									
    //host adatainak nullazasa
    bzero(&host_addr,sizeof(host_addr));
    //host IP cimet atmasoljuk a host adatok koze
    bcopy(host_entry->h_addr_list[0],
	    &host_addr.sin_addr.s_addr,
	     host_entry->h_length);

    //meg kell adni, hogy milyen tipusu a cim. AF_INET lesz
    host_addr.sin_family = host_entry->h_addrtype;
    //megadjuk a portot. htons -al forditani kell a byte sorrenden
    host_addr.sin_port = htons(port);
						  
    //socket nonblocking lesz
    connect(sd, (struct sockaddr *)&host_addr, sizeof(struct sockaddr_in));

    return sd;
}


void main(int argc, char *argv[])
{
    int port;
    
    if (argc!=2) exit();
    //lookup
    if ((host_entry = gethostbyname(argv[1]))==NULL)
    {
	herror("gethostbyname failed");
	exit(1);
    }

    for(port=0;port<65536;port+=MAXSOCKETS+1) 
	scanports(port,port+MAXSOCKETS<65535 ? port+MAXSOCKETS : 65535);
}